/*
 * bsc1205130_fs_nfsd_nfssvc
 *
 * Fix for CVE-2022-43945, bsc#1205130
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/nfsd/nfssvc.c */
#include <linux/sched/signal.h>
#include <linux/module.h>
#include <linux/fs_struct.h>
#include <linux/swap.h>
#include <linux/sunrpc/svcsock.h>

/* klp-ccp: from include/linux/sunrpc/xdr.h */
static void (*klpe_xdr_init_decode)(struct xdr_stream *xdr, struct xdr_buf *buf,
			    __be32 *p, struct rpc_rqst *rqst);

/* klp-ccp: from fs/nfsd/nfssvc.c */
#include <linux/sunrpc/svc_xprt.h>
#include <linux/seq_file.h>
#include <net/ipv6.h>
#include <net/net_namespace.h>
/* klp-ccp: from fs/nfsd/nfsd.h */
#include <linux/types.h>
#include <linux/mount.h>
#include <linux/nfs.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/svc_xprt.h>
#include <linux/sunrpc/msg_prot.h>
#include <uapi/linux/nfsd/debug.h>
/* klp-ccp: from fs/nfsd/netns.h */
#include <net/net_namespace.h>
#include <linux/percpu_counter.h>
/* klp-ccp: from fs/nfsd/export.h */
#include <linux/sunrpc/cache.h>
#include <linux/percpu_counter.h>
#include <uapi/linux/nfsd/export.h>
#include <linux/nfs4.h>
/* klp-ccp: from fs/nfsd/stats.h */
#include <uapi/linux/nfsd/stats.h>
#include <linux/percpu_counter.h>

/* klp-ccp: from fs/nfsd/cache.h */
#include <linux/sunrpc/svc.h>

enum {
	RC_DROPIT,
	RC_REPLY,
	RC_DOIT
};

enum {
	RC_NOCACHE,
	RC_REPLSTAT,
	RC_REPLBUFF,
};

static int	(*klpe_nfsd_cache_lookup)(struct svc_rqst *);
static void	(*klpe_nfsd_cache_update)(struct svc_rqst *, int, __be32 *);

/* klp-ccp: from fs/nfsd/nfsfh.h */
#include <linux/crc32.h>
#include <linux/sunrpc/svc.h>

#include "common.h"

/* klp-ccp: from fs/nfsd/trace.h */
KLPR_TRACE_EVENT(nfsd_garbage_args_err,
	TP_PROTO(const struct svc_rqst *rqst),
	TP_ARGS(rqst)
);

KLPR_TRACE_EVENT(nfsd_cant_encode_err,
	TP_PROTO(const struct svc_rqst *rqst),
	TP_ARGS(rqst)
);

/* klp-ccp: from fs/nfsd/state.h */
#include <linux/idr.h>
#include <linux/refcount.h>
#include <linux/sunrpc/svc_xprt.h>

#include <trace/define_trace.h>

#include "svc.h"

/* klp-ccp: from fs/nfsd/nfssvc.c */
int klpp_nfsd_dispatch(struct svc_rqst *rqstp, __be32 *statp)
{
	const struct svc_procedure *proc = rqstp->rq_procinfo;
	struct kvec *argv = &rqstp->rq_arg.head[0];
	struct kvec *resv = &rqstp->rq_res.head[0];
	__be32 *p;

	/*
	 * Give the xdr decoder a chance to change this if it wants
	 * (necessary in the NFSv4.0 compound case)
	 */
	rqstp->rq_cachetype = proc->pc_cachetype;

	klpp_svcxdr_init_decode(rqstp);
	if (!proc->pc_decode(rqstp, argv->iov_base))
		goto out_decode_err;

	switch ((*klpe_nfsd_cache_lookup)(rqstp)) {
	case RC_DOIT:
		break;
	case RC_REPLY:
		goto out_cached_reply;
	case RC_DROPIT:
		goto out_dropit;
	}

	/*
	 * Need to grab the location to store the status, as
	 * NFSv4 does some encoding while processing
	 */
	p = resv->iov_base + resv->iov_len;
	klpp_svcxdr_init_encode(rqstp);

	*statp = proc->pc_func(rqstp);
	if (*statp == rpc_drop_reply || test_bit(RQ_DROPME, &rqstp->rq_flags))
		goto out_update_drop;

	if (!proc->pc_encode(rqstp, p))
		goto out_encode_err;

	(*klpe_nfsd_cache_update)(rqstp, rqstp->rq_cachetype, statp + 1);
out_cached_reply:
	return 1;

out_decode_err:
	klpr_trace_nfsd_garbage_args_err(rqstp);
	*statp = rpc_garbage_args;
	return 1;

out_update_drop:
	(*klpe_nfsd_cache_update)(rqstp, RC_NOCACHE, NULL);
out_dropit:
	return 0;

out_encode_err:
	klpr_trace_nfsd_cant_encode_err(rqstp);
	(*klpe_nfsd_cache_update)(rqstp, RC_NOCACHE, NULL);
	*statp = rpc_system_err;
	return 1;
}



#define LP_MODULE "nfsd"

#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1205130.h"
#include "../kallsyms_relocs.h"



static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__traceiter_nfsd_cant_encode_err",
	  (void *)&klpe___traceiter_nfsd_cant_encode_err, "nfsd" },
	{ "__tracepoint_nfsd_cant_encode_err",
	  (void *)&klpe___tracepoint_nfsd_cant_encode_err, "nfsd" },
	{ "__traceiter_nfsd_garbage_args_err",
	  (void *)&klpe___traceiter_nfsd_garbage_args_err, "nfsd" },
	{ "__tracepoint_nfsd_garbage_args_err",
	  (void *)&klpe___tracepoint_nfsd_garbage_args_err, "nfsd" },
	{ "nfsd_cache_lookup", (void *)&klpe_nfsd_cache_lookup, "nfsd" },
	{ "nfsd_cache_update", (void *)&klpe_nfsd_cache_update, "nfsd" },
	{ "xdr_init_decode", (void *)&klpe_xdr_init_decode, "sunrpc" },
};

static int bsc1205130_fs_nfsd_nfssvc_module_notify(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LP_MODULE))
		return 0;
	ret = klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));

	WARN(ret, "%s: delayed kallsyms lookup failed. System is broken and can crash.\n",
		__func__);

	return ret;
}

static struct notifier_block module_nb = {
	.notifier_call = bsc1205130_fs_nfsd_nfssvc_module_notify,
	.priority = INT_MIN+1,
};

int bsc1205130_fs_nfsd_nfssvc_init(void)
{
	int ret;
	struct module *mod;

	ret = klp_kallsyms_relocs_init();
	if (ret)
		return ret;

	ret = register_module_notifier(&module_nb);
	if (ret)
		return ret;

	rcu_read_lock_sched();
	mod = (*klpe_find_module)(LP_MODULE);
	if (!try_module_get(mod))
		mod = NULL;
	rcu_read_unlock_sched();

	if (mod) {
		ret = klp_resolve_kallsyms_relocs(klp_funcs,
						ARRAY_SIZE(klp_funcs));
	}

	if (ret)
		unregister_module_notifier(&module_nb);
	module_put(mod);

	return ret;
}

void bsc1205130_fs_nfsd_nfssvc_cleanup(void)
{
	unregister_module_notifier(&module_nb);
}
